// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: pairing.proto
// Protobuf C++ Version: 7.35.1

#ifndef pairing_2eproto_2epb_2eh
#define pairing_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_bases.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/generated_enum_reflection.h"
#include "google/protobuf/unknown_field_set.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_pairing_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_pairing_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_pairing_2eproto;
}  // extern "C"
namespace adb {
namespace proto {
enum PairingPacket_Type : int;
extern const uint32_t PairingPacket_Type_internal_data_[];
class PairingPacket;
struct PairingPacketGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern PairingPacketGlobalsTypeInternal PairingPacket_globals_;
extern const ::google::protobuf::internal::ClassDataFull PairingPacket_class_data_;
#else
extern const PairingPacketGlobalsTypeInternal PairingPacket_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace proto
}  // namespace adb
namespace google {
namespace protobuf {
template <>
internal::EnumTraitsT<::adb::proto::PairingPacket_Type_internal_data_>
    internal::EnumTraitsImpl::value<::adb::proto::PairingPacket_Type>;
}  // namespace protobuf
}  // namespace google

namespace adb {
namespace proto {
enum PairingPacket_Type : int {
  PairingPacket_Type_SPAKE2_MSG = 0,
  PairingPacket_Type_PEER_INFO = 1,
  PairingPacket_Type_PairingPacket_Type_INT_MIN_SENTINEL_DO_NOT_USE_ =
      ::std::numeric_limits<::int32_t>::min(),
  PairingPacket_Type_PairingPacket_Type_INT_MAX_SENTINEL_DO_NOT_USE_ =
      ::std::numeric_limits<::int32_t>::max(),
};

extern const uint32_t PairingPacket_Type_internal_data_[];
inline constexpr PairingPacket_Type PairingPacket_Type_Type_MIN =
    static_cast<PairingPacket_Type>(0);
inline constexpr PairingPacket_Type PairingPacket_Type_Type_MAX =
    static_cast<PairingPacket_Type>(1);
[[nodiscard]] inline bool PairingPacket_Type_IsValid(int value) {
  return 0 <= value && value <= 1;
}
inline constexpr int PairingPacket_Type_Type_ARRAYSIZE = 1 + 1;
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
PairingPacket_Type_descriptor();
[[nodiscard]] inline auto ProtobufInternalGetEnumDescriptor(PairingPacket_Type) {
  return PairingPacket_Type_descriptor();
}
template <typename T>
[[nodiscard]] const ::std::string& PairingPacket_Type_Name(T value) {
  static_assert(::std::is_same<T, PairingPacket_Type>::value ||
                    ::std::is_integral<T>::value,
                "Incorrect type passed to Type_Name().");
  return PairingPacket_Type_Name(static_cast<PairingPacket_Type>(value));
}
template <>
[[nodiscard]] inline const ::std::string& PairingPacket_Type_Name(PairingPacket_Type value) {
  return ::google::protobuf::internal::NameOfDenseEnum<PairingPacket_Type_descriptor, 0, 1>(
      static_cast<int>(value));
}
[[nodiscard]] inline bool PairingPacket_Type_Parse(
    ::absl::string_view name, PairingPacket_Type* PROTOBUF_NONNULL value) {
  return ::google::protobuf::internal::ParseNamedEnum<PairingPacket_Type>(PairingPacket_Type_descriptor(), name,
                                           value);
}
using ::google::protobuf::internal::generated_enum::AbslParseFlag;
using ::google::protobuf::internal::generated_enum::AbslUnparseFlag;

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED PairingPacket final : public ::google::protobuf::internal::ZeroFieldsBase
/* @@protoc_insertion_point(class_definition:adb.proto.PairingPacket) */ {
 public:
  inline PairingPacket() : PairingPacket(nullptr) {}

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(PairingPacket* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(PairingPacket));
  }
#endif

  template <typename = void>
  explicit constexpr PairingPacket(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline PairingPacket(const PairingPacket& from) : PairingPacket(nullptr, from) {}
  inline PairingPacket(PairingPacket&& from) noexcept : PairingPacket(nullptr, ::std::move(from)) {}
  inline PairingPacket& operator=(const PairingPacket& from) {
    CopyFrom(from);
    return *this;
  }
  inline PairingPacket& operator=(PairingPacket&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const PairingPacket& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<PairingPacket>(&PairingPacket_globals_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(PairingPacket& a, PairingPacket& b) { a.Swap(&b); }
  inline void Swap(PairingPacket* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(PairingPacket* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] PairingPacket* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::internal::ZeroFieldsBase::DefaultConstruct<PairingPacket>(arena);
  }
  using ::google::protobuf::internal::ZeroFieldsBase::CopyFrom;
  inline void CopyFrom(const PairingPacket& from) { ::google::protobuf::internal::ZeroFieldsBase::CopyImpl(*this, from); }
  using ::google::protobuf::internal::ZeroFieldsBase::MergeFrom;
  void MergeFrom(const PairingPacket& from) { ::google::protobuf::internal::ZeroFieldsBase::MergeImpl(*this, from); }

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "adb.proto.PairingPacket"; }

  explicit PairingPacket(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  PairingPacket(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const PairingPacket& from);
  PairingPacket(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, PairingPacket&& from) noexcept
      : PairingPacket(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------
  using Type = PairingPacket_Type;
  static constexpr Type SPAKE2_MSG = PairingPacket_Type_SPAKE2_MSG;
  static constexpr Type PEER_INFO = PairingPacket_Type_PEER_INFO;
  [[nodiscard]] static inline bool Type_IsValid(int value) {
    return PairingPacket_Type_IsValid(value);
  }
  static constexpr Type Type_MIN = PairingPacket_Type_Type_MIN;
  static constexpr Type Type_MAX = PairingPacket_Type_Type_MAX;
  static constexpr int Type_ARRAYSIZE = PairingPacket_Type_Type_ARRAYSIZE;
  static inline const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL Type_descriptor() {
    return PairingPacket_Type_descriptor();
  }
  template <typename T>
  [[nodiscard]] static inline const ::std::string& Type_Name(T value) {
    return PairingPacket_Type_Name(value);
  }
  [[nodiscard]] static inline bool Type_Parse(
      ::absl::string_view name, Type* PROTOBUF_NONNULL value) {
    return PairingPacket_Type_Parse(name, value);
  }

  // accessors -------------------------------------------------------
  // @@protoc_insertion_point(class_scope:adb.proto.PairingPacket)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<0, 0,
                          0, 0,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  friend struct ::TableStruct_pairing_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// PairingPacket

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace proto
}  // namespace adb


namespace google {
namespace protobuf {

template <>
struct is_proto_enum<::adb::proto::PairingPacket_Type> : std::true_type {};
template <>
inline const EnumDescriptor* PROTOBUF_NONNULL GetEnumDescriptor<::adb::proto::PairingPacket_Type>() {
  return ::adb::proto::PairingPacket_Type_descriptor();
}

}  // namespace protobuf
}  // namespace google

// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // pairing_2eproto_2epb_2eh
