// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: app_processes.proto
// Protobuf C++ Version: 7.35.1

#ifndef app_5fprocesses_2eproto_2epb_2eh
#define app_5fprocesses_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/unknown_field_set.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_app_5fprocesses_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_app_5fprocesses_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_app_5fprocesses_2eproto;
}  // extern "C"
namespace adb {
namespace proto {
class AppProcesses;
struct AppProcessesGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern AppProcessesGlobalsTypeInternal AppProcesses_globals_;
extern const ::google::protobuf::internal::ClassDataFull AppProcesses_class_data_;
#else
extern const AppProcessesGlobalsTypeInternal AppProcesses_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class ProcessEntry;
struct ProcessEntryGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern ProcessEntryGlobalsTypeInternal ProcessEntry_globals_;
extern const ::google::protobuf::internal::ClassDataFull ProcessEntry_class_data_;
#else
extern const ProcessEntryGlobalsTypeInternal ProcessEntry_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace proto
}  // namespace adb
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace adb {
namespace proto {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED ProcessEntry final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:adb.proto.ProcessEntry) */ {
 public:
  inline ProcessEntry() : ProcessEntry(nullptr) {}
  ~ProcessEntry() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(ProcessEntry* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(ProcessEntry));
  }
#endif

  template <typename = void>
  explicit constexpr ProcessEntry(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline ProcessEntry(const ProcessEntry& from) : ProcessEntry(nullptr, from) {}
  inline ProcessEntry(ProcessEntry&& from) noexcept : ProcessEntry(nullptr, ::std::move(from)) {}
  inline ProcessEntry& operator=(const ProcessEntry& from) {
    CopyFrom(from);
    return *this;
  }
  inline ProcessEntry& operator=(ProcessEntry&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const ProcessEntry& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<ProcessEntry>(&ProcessEntry_globals_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(ProcessEntry& a, ProcessEntry& b) { a.Swap(&b); }
  inline void Swap(ProcessEntry* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(ProcessEntry* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] ProcessEntry* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<ProcessEntry>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const ProcessEntry& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const ProcessEntry& from) { ProcessEntry::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(ProcessEntry* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "adb.proto.ProcessEntry"; }

  explicit ProcessEntry(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  ProcessEntry(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ProcessEntry& from);
  ProcessEntry(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, ProcessEntry&& from) noexcept
      : ProcessEntry(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kPackageNamesFieldNumber = 7,
    kArchitectureFieldNumber = 4,
    kProcessNameFieldNumber = 6,
    kPidFieldNumber = 1,
    kUserIdFieldNumber = 5,
    kUidFieldNumber = 9,
    kDebuggableFieldNumber = 2,
    kProfileableFieldNumber = 3,
    kWaitingForDebuggerFieldNumber = 8,
  };
  // repeated string package_names = 7;
  [[nodiscard]] int package_names_size()
      const;
  private:
  int _internal_package_names_size() const;

  public:
  void clear_package_names() ;
  [[nodiscard]] const ::std::string& package_names(int index) const;
  ::std::string* PROTOBUF_NONNULL mutable_package_names(int index);
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_package_names(int index, Arg_&& value, Args_... args);
  ::std::string* PROTOBUF_NONNULL add_package_names();
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void add_package_names(Arg_&& value, Args_... args);
  [[nodiscard]] const
      ::google::protobuf::RepeatedPtrField<::std::string>&
      package_names() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::std::string>*
      PROTOBUF_NONNULL
      mutable_package_names();

  private:
  const ::google::protobuf::RepeatedPtrField<::std::string>& _internal_package_names() const;
  ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL _internal_mutable_package_names();

  public:
  // string architecture = 4;
  void clear_architecture() ;
  [[nodiscard]] const ::std::string& architecture() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_architecture(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_architecture();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_architecture();
  void set_allocated_architecture(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_architecture() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_architecture(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_architecture();

  public:
  // optional string process_name = 6;
  [[nodiscard]] bool has_process_name()
      const;
  void clear_process_name() ;
  [[nodiscard]] const ::std::string& process_name() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_process_name(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_process_name();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_process_name();
  void set_allocated_process_name(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_process_name() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_process_name(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_process_name();

  public:
  // int64 pid = 1;
  void clear_pid() ;
  [[nodiscard]] ::int64_t pid() const;
  void set_pid(::int64_t value);

  private:
  ::int64_t _internal_pid() const;
  void _internal_set_pid(::int64_t value);

  public:
  // optional int64 user_id = 5;
  [[nodiscard]] bool has_user_id()
      const;
  void clear_user_id() ;
  [[nodiscard]] ::int64_t user_id() const;
  void set_user_id(::int64_t value);

  private:
  ::int64_t _internal_user_id() const;
  void _internal_set_user_id(::int64_t value);

  public:
  // optional int64 uid = 9;
  [[nodiscard]] bool has_uid()
      const;
  void clear_uid() ;
  [[nodiscard]] ::int64_t uid() const;
  void set_uid(::int64_t value);

  private:
  ::int64_t _internal_uid() const;
  void _internal_set_uid(::int64_t value);

  public:
  // bool debuggable = 2;
  void clear_debuggable() ;
  [[nodiscard]] bool debuggable() const;
  void set_debuggable(bool value);

  private:
  bool _internal_debuggable() const;
  void _internal_set_debuggable(bool value);

  public:
  // bool profileable = 3;
  void clear_profileable() ;
  [[nodiscard]] bool profileable() const;
  void set_profileable(bool value);

  private:
  bool _internal_profileable() const;
  void _internal_set_profileable(bool value);

  public:
  // optional bool waiting_for_debugger = 8;
  [[nodiscard]] bool has_waiting_for_debugger()
      const;
  void clear_waiting_for_debugger() ;
  [[nodiscard]] bool waiting_for_debugger() const;
  void set_waiting_for_debugger(bool value);

  private:
  bool _internal_waiting_for_debugger() const;
  void _internal_set_waiting_for_debugger(bool value);

  public:
  // @@protoc_insertion_point(class_scope:adb.proto.ProcessEntry)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<4, 9,
                          0, 76,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const ProcessEntry& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField<::std::string> package_names_;
    ::google::protobuf::internal::ArenaStringPtr architecture_;
    ::google::protobuf::internal::ArenaStringPtr process_name_;
    ::int64_t pid_;
    ::int64_t user_id_;
    ::int64_t uid_;
    bool debuggable_;
    bool profileable_;
    bool waiting_for_debugger_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_app_5fprocesses_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED AppProcesses final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:adb.proto.AppProcesses) */ {
 public:
  inline AppProcesses() : AppProcesses(nullptr) {}
  ~AppProcesses() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(AppProcesses* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(AppProcesses));
  }
#endif

  template <typename = void>
  explicit constexpr AppProcesses(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline AppProcesses(const AppProcesses& from) : AppProcesses(nullptr, from) {}
  inline AppProcesses(AppProcesses&& from) noexcept : AppProcesses(nullptr, ::std::move(from)) {}
  inline AppProcesses& operator=(const AppProcesses& from) {
    CopyFrom(from);
    return *this;
  }
  inline AppProcesses& operator=(AppProcesses&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const AppProcesses& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<AppProcesses>(&AppProcesses_globals_);
  }
  static constexpr int kIndexInFileMessages = 1;
  friend void swap(AppProcesses& a, AppProcesses& b) { a.Swap(&b); }
  inline void Swap(AppProcesses* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(AppProcesses* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] AppProcesses* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<AppProcesses>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const AppProcesses& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const AppProcesses& from) { AppProcesses::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(AppProcesses* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "adb.proto.AppProcesses"; }

  explicit AppProcesses(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  AppProcesses(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const AppProcesses& from);
  AppProcesses(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, AppProcesses&& from) noexcept
      : AppProcesses(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kProcessFieldNumber = 1,
  };
  // repeated .adb.proto.ProcessEntry process = 1;
  [[nodiscard]] int process_size()
      const;
  private:
  int _internal_process_size() const;

  public:
  void clear_process() ;
  [[nodiscard]] const ::adb::proto::ProcessEntry& process(int index) const;
  [[nodiscard]] ::adb::proto::ProcessEntry* PROTOBUF_NONNULL mutable_process(int index);
  ::adb::proto::ProcessEntry* PROTOBUF_NONNULL add_process();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>&
  process() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>* PROTOBUF_NONNULL
  mutable_process();

  private:
  const ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>& _internal_process() const;
  ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>* PROTOBUF_NONNULL _internal_mutable_process();

  public:
  // @@protoc_insertion_point(class_scope:adb.proto.AppProcesses)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<0, 1,
                          1, 0,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const AppProcesses& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField< ::adb::proto::ProcessEntry > process_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_app_5fprocesses_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// ProcessEntry

// int64 pid = 1;
inline void ProcessEntry::clear_pid() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.pid_ = ::int64_t{0};
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline ::int64_t ProcessEntry::pid() const {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.pid)
  return _internal_pid();
}
inline void ProcessEntry::set_pid(::int64_t value) {
  _internal_set_pid(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.pid)
}
inline ::int64_t ProcessEntry::_internal_pid() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.pid_;
}
inline void ProcessEntry::_internal_set_pid(::int64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.pid_ = value;
}

// bool debuggable = 2;
inline void ProcessEntry::clear_debuggable() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.debuggable_ = false;
  ClearHasBit(_impl_._has_bits_[0], 0x00000040U);
}
inline bool ProcessEntry::debuggable() const {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.debuggable)
  return _internal_debuggable();
}
inline void ProcessEntry::set_debuggable(bool value) {
  _internal_set_debuggable(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000040U);
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.debuggable)
}
inline bool ProcessEntry::_internal_debuggable() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.debuggable_;
}
inline void ProcessEntry::_internal_set_debuggable(bool value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.debuggable_ = value;
}

// bool profileable = 3;
inline void ProcessEntry::clear_profileable() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.profileable_ = false;
  ClearHasBit(_impl_._has_bits_[0], 0x00000080U);
}
inline bool ProcessEntry::profileable() const {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.profileable)
  return _internal_profileable();
}
inline void ProcessEntry::set_profileable(bool value) {
  _internal_set_profileable(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000080U);
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.profileable)
}
inline bool ProcessEntry::_internal_profileable() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.profileable_;
}
inline void ProcessEntry::_internal_set_profileable(bool value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.profileable_ = value;
}

// string architecture = 4;
inline void ProcessEntry::clear_architecture() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.architecture_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& ProcessEntry::architecture() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.architecture)
  return _internal_architecture();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ProcessEntry::set_architecture(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.architecture_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.architecture)
}
inline ::std::string* PROTOBUF_NONNULL ProcessEntry::mutable_architecture()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_architecture();
  // @@protoc_insertion_point(field_mutable:adb.proto.ProcessEntry.architecture)
  return _s;
}
inline const ::std::string& ProcessEntry::_internal_architecture() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.architecture_.Get();
}
inline void ProcessEntry::_internal_set_architecture(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.architecture_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ProcessEntry::_internal_mutable_architecture() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.architecture_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ProcessEntry::release_architecture() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:adb.proto.ProcessEntry.architecture)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.architecture_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.architecture_.Set("", GetArena());
  }
  return released;
}
inline void ProcessEntry::set_allocated_architecture(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.architecture_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.architecture_.IsDefault()) {
    _impl_.architecture_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:adb.proto.ProcessEntry.architecture)
}

// optional int64 user_id = 5;
inline bool ProcessEntry::has_user_id() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000010U);
  return value;
}
inline void ProcessEntry::clear_user_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.user_id_ = ::int64_t{0};
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline ::int64_t ProcessEntry::user_id() const {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.user_id)
  return _internal_user_id();
}
inline void ProcessEntry::set_user_id(::int64_t value) {
  _internal_set_user_id(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.user_id)
}
inline ::int64_t ProcessEntry::_internal_user_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.user_id_;
}
inline void ProcessEntry::_internal_set_user_id(::int64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.user_id_ = value;
}

// optional string process_name = 6;
inline bool ProcessEntry::has_process_name() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000004U);
  return value;
}
inline void ProcessEntry::clear_process_name() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.process_name_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::std::string& ProcessEntry::process_name() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.process_name)
  return _internal_process_name();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ProcessEntry::set_process_name(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  _impl_.process_name_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.process_name)
}
inline ::std::string* PROTOBUF_NONNULL ProcessEntry::mutable_process_name()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::std::string* _s = _internal_mutable_process_name();
  // @@protoc_insertion_point(field_mutable:adb.proto.ProcessEntry.process_name)
  return _s;
}
inline const ::std::string& ProcessEntry::_internal_process_name() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.process_name_.Get();
}
inline void ProcessEntry::_internal_set_process_name(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.process_name_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ProcessEntry::_internal_mutable_process_name() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.process_name_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ProcessEntry::release_process_name() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:adb.proto.ProcessEntry.process_name)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000004U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  auto* released = _impl_.process_name_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.process_name_.Set("", GetArena());
  }
  return released;
}
inline void ProcessEntry::set_allocated_process_name(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  _impl_.process_name_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.process_name_.IsDefault()) {
    _impl_.process_name_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:adb.proto.ProcessEntry.process_name)
}

// repeated string package_names = 7;
inline int ProcessEntry::_internal_package_names_size() const {
  return _internal_package_names().size();
}
inline int ProcessEntry::package_names_size() const {
  return _internal_package_names_size();
}
inline void ProcessEntry::clear_package_names() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.package_names_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline ::std::string* PROTOBUF_NONNULL ProcessEntry::add_package_names()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::std::string* _s =
      _internal_mutable_package_names()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add_mutable:adb.proto.ProcessEntry.package_names)
  return _s;
}
inline const ::std::string& ProcessEntry::package_names(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.package_names)
  return _internal_package_names().Get(index);
}
inline ::std::string* PROTOBUF_NONNULL ProcessEntry::mutable_package_names(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:adb.proto.ProcessEntry.package_names)
  return _internal_mutable_package_names()->Mutable(index);
}
template <typename Arg_, typename... Args_>
inline void ProcessEntry::set_package_names(int index, Arg_&& value, Args_... args) {
  ::google::protobuf::internal::AssignToString(
      *_internal_mutable_package_names()->Mutable(index),
      ::std::forward<Arg_>(value), args... );
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.package_names)
}
template <typename Arg_, typename... Args_>
inline void ProcessEntry::add_package_names(Arg_&& value, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::google::protobuf::internal::AddToRepeatedPtrField(
      ::google::protobuf::MessageLite::internal_visibility(), GetArena(),
      *_internal_mutable_package_names(), ::std::forward<Arg_>(value),
      args... );
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:adb.proto.ProcessEntry.package_names)
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>& ProcessEntry::package_names()
    const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:adb.proto.ProcessEntry.package_names)
  return _internal_package_names();
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
ProcessEntry::mutable_package_names() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:adb.proto.ProcessEntry.package_names)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_package_names();
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>&
ProcessEntry::_internal_package_names() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.package_names_;
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
ProcessEntry::_internal_mutable_package_names() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.package_names_;
}

// optional bool waiting_for_debugger = 8;
inline bool ProcessEntry::has_waiting_for_debugger() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000100U);
  return value;
}
inline void ProcessEntry::clear_waiting_for_debugger() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.waiting_for_debugger_ = false;
  ClearHasBit(_impl_._has_bits_[0], 0x00000100U);
}
inline bool ProcessEntry::waiting_for_debugger() const {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.waiting_for_debugger)
  return _internal_waiting_for_debugger();
}
inline void ProcessEntry::set_waiting_for_debugger(bool value) {
  _internal_set_waiting_for_debugger(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000100U);
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.waiting_for_debugger)
}
inline bool ProcessEntry::_internal_waiting_for_debugger() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.waiting_for_debugger_;
}
inline void ProcessEntry::_internal_set_waiting_for_debugger(bool value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.waiting_for_debugger_ = value;
}

// optional int64 uid = 9;
inline bool ProcessEntry::has_uid() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000020U);
  return value;
}
inline void ProcessEntry::clear_uid() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.uid_ = ::int64_t{0};
  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
}
inline ::int64_t ProcessEntry::uid() const {
  // @@protoc_insertion_point(field_get:adb.proto.ProcessEntry.uid)
  return _internal_uid();
}
inline void ProcessEntry::set_uid(::int64_t value) {
  _internal_set_uid(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  // @@protoc_insertion_point(field_set:adb.proto.ProcessEntry.uid)
}
inline ::int64_t ProcessEntry::_internal_uid() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.uid_;
}
inline void ProcessEntry::_internal_set_uid(::int64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.uid_ = value;
}

// -------------------------------------------------------------------

// AppProcesses

// repeated .adb.proto.ProcessEntry process = 1;
inline int AppProcesses::_internal_process_size() const {
  return _internal_process().size();
}
inline int AppProcesses::process_size() const {
  return _internal_process_size();
}
inline void AppProcesses::clear_process() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.process_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::adb::proto::ProcessEntry& AppProcesses::process(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:adb.proto.AppProcesses.process)
  return _internal_process().Get(index);
}
inline ::adb::proto::ProcessEntry* PROTOBUF_NONNULL AppProcesses::mutable_process(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:adb.proto.AppProcesses.process)
  return _internal_mutable_process()->Mutable(index);
}
inline ::adb::proto::ProcessEntry* PROTOBUF_NONNULL AppProcesses::add_process()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::adb::proto::ProcessEntry* _add =
      _internal_mutable_process()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:adb.proto.AppProcesses.process)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>& AppProcesses::process() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:adb.proto.AppProcesses.process)
  return _internal_process();
}
inline ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>* PROTOBUF_NONNULL
AppProcesses::mutable_process() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:adb.proto.AppProcesses.process)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_process();
}
inline const ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>&
AppProcesses::_internal_process() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.process_;
}
inline ::google::protobuf::RepeatedPtrField<::adb::proto::ProcessEntry>* PROTOBUF_NONNULL
AppProcesses::_internal_mutable_process() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.process_;
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace proto
}  // namespace adb


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // app_5fprocesses_2eproto_2epb_2eh
